/* export.c */

#include "df3util.h"

static char const * const TU = "export";

/* globals */
static char *asrc,      /* source name arg, null if default */
            *atgt;      /* target (png) name arg */
static char aaxis;      /* axis arg */
static size_t aoffs;    /* n arg */

static df3Axes saxes;   /* source axes arg, if source default */
static size_t sbpv;     /* source bpv arg, if source default */

static df3Handle df3i;

static char *pngname;   /* target name alloc'd */
static uint8_t *pbuf;   /* alloc'd.  the image data */
                        /* nb bigendian when bpv=2 */
static size_t icols,    /* image dims */
              irows;



/* procArgs()
 * -------------------------------------------------------------------------- *
 * cli := source axis:n target.
 *     := . axes bpv axis:n target.
 */

static bool procArgs(char *av[]) {
  bool f = false;
  char buf[PATH_MAX];
  int const rc = getInpSrc(av, &saxes, &sbpv);
  int i;

  if (-1 == rc)
    goto label_fail;
  else if (1 == rc) {
    asrc = NULL;  /* redundant, but hey.. */
    i = 3;
  }
  else if (!getArgFname(av[0], buf)) {
    EMITEMSG(false, "bad source file name");
    goto label_fail;
  }
  else {
    asrc = av[0];
    i = 1;
  }

  if (!(av[i]  &&  av[i+1])) {
    EMITEMSG(false, "missing argument(s)");
    app_lastec = ec_badUsage;
    goto label_fail;
  }

  int n = 0;
  if (!getArgAxisN(av[i], &aaxis, &n)) {
    EMITEMSG(false, "bad axis:n argument");
    goto label_fail;
  }
  else if (0 > n) {
    app_lastec = ec_badArgument;
    EMITEMSG(false, "negative axis:n argument");
    goto label_fail;
  }
  aoffs = (size_t)n;

  if (!getArgFname(av[i+1], buf)) {
    EMITEMSG(false, "bad target file name");
    goto label_fail;
  }
  atgt = av[i+1];

  f = true;

label_fail:
  return f;
}


/* openSource()
 * -------------------------------------------------------------------------- *
 * open if actual file, verify comaptible bpv + aoffs in range.
 */


static bool openSource(void) {

  if (asrc) {
    df3Error ec;
    if (!mkSourceFname(asrc, false, false, &df3i.fname)) {
      EMITEMSG(false, "source '%s': %s",
              trimmedFn(asrc), appLastError(app_lastec));
      return false;
    }
    else if (ecOk != (ec = df3_open_file(&df3i, true))) {
      EMITEMSG(false, "open source '%s': %s",
              trimmedFn(df3i.fname), df3_error_message(ec));
      app_lastec = le2ae(ec);
      return false;
    }
  }
  else {
    df3i.axes = saxes;
    df3i.bpv = sbpv;
  }

  if (app_diagnose) {
    if (asrc) {
      EMITDMSG(TU, "source handle");
      df3_handle_info(&df3i, app_logfp);
    }
    else
      EMITDMSG(TU, "source handle <stdin>");
  }

  bool f = false;
  if (4 == df3i.bpv) {
    EMITEMSG(false, "source bpv is 4");
    app_lastec = ec_badArgument;
  }
  else if (badAcol(df3i.axes, aaxis, aoffs)) {
    EMITEMSG(false, "axis '%c' range 0,..,%zu",
            aaxis, axisMax(df3i.axes, aaxis) - 1);
    app_lastec = ec_badArgument;
  }
  else
    f = true;

  if (!f)
    (void)df3_close_file(&df3i);

  return f;
}


/* prepTargetBuffer()
 * -------------------------------------------------------------------------- *
 * (over)estimate image size by doubling raw data, -fill if stdin data.
 */

static bool prepTargetBuffer(void) {
  char const * const fmt = "target image '%s': %s";
  mkBufferXY(df3i.axes, aaxis, &icols, &irows);
  if (!haveDiskSpace((df3Axes){icols,irows,2}, df3i.bpv)) {
    EMITEMSG(false, fmt, trimmedFn(atgt),
            appLastError(app_lastec = ec_diskFull));
    return false;
  }
  else if (!mkTargetFname(atgt, true, &pngname)) {
    EMITEMSG(false, fmt, trimmedFn(atgt), appLastError(app_lastec));
    return false;
  }
  else if (!(pbuf = calloc(icols * irows * df3i.bpv, sizeof(uint8_t)))) {
    int const e = errno;
    EMITEMSG(false, "creating export buffer: %s", strerror(e));
    app_lastec = ec_sysError;
    return false;
  }
// TODO prepFill buffer if default source.

  return true;
}


/* closeFile()
 * -------------------------------------------------------------------------- *
 * TODO what's lost by setting {app_lastec} here?  error toggle {f}?
 * TODO use EMITEMSGs instead + unconditionally?
 */

static bool closeFile(void) {
  bool f = true;

  if (asrc) {
    df3Error const ec = df3_close_file(&df3i);
    if (ecOk != ec)
      if (app_diagnose)
        EMITDMSG(TU, "close source: %s", df3_error_message(ec));
    if (app_diagnose)
      EMITDMSG(TU, "closed source handle: %s",
              df3_handle_error(&df3i, false));
    if (ecOk != ec   &&  ec_sysError != app_lastec)
      app_lastec = le2ae(ec);
  }
  else {
    (void)fclose(stdin);
    if (app_diagnose)
      EMITDMSG(TU, "closed source <stdin>");
  }

  return f;
}


/* buffer2png()
 * -------------------------------------------------------------------------- *
 * use fileLoop or callback to fill buffer.
 * nb bigendian for bpv=2.
 */

static bool cback(void *vp, df3Coord const at, uint32_t const v) {
  df3Handle * const df3 = vp;

  if (!inpVoxelValid(df3->axes, df3->bpv, at, v))
    return false;

  bool todo = false;
  size_t x, y;
  switch (aaxis) {
    case 'x':
      if (at.x == aoffs) {
        x = at.z;
        y = at.y;
        todo = true;
      }
      break;
    case 'y':
      if (at.y == aoffs) {
        x = at.x;
        y = at.z;
        todo = true;
      }
      break;
    case 'z':
      if (at.z == aoffs) {
        x = at.x;
        y = at.y;
        todo = true;
      }
      break;
    default:
      APPPANIC;
  }

  if (todo) {
    if (2 == df3->bpv) {
      uint8_t * const p = pbuf + 2 * icols * y + 2 * x;
      *p = (uint8_t)(v / 256);  *(p + 1) = (uint8_t)(v % 256);
    }
    else
      *(pbuf + icols * y + x) = (uint8_t)v;
  }

  return true;
}

static bool fileLoop(void) {
  for (size_t y = 0; irows != y; ++y)
    for (size_t x = 0; icols != x; ++x) {
      df3Coord at;
      switch (aaxis) {
        case 'x':  at.x = aoffs;  at.y = y;  at.z = x;  break;
        case 'y':  at.x = x;  at.y = aoffs;  at.z = y;  break;
        case 'z':  at.x = x;  at.y = y;  at.z = aoffs;  break;
        default:   APPPANIC;
      }
      uint32_t v;
      df3Error ec = df3_get_voxel(&df3i, at, &v);
      if (ecOk != ec) {
        EMITEMSG(false, "reading voxel <%zu,%zu,%zu>: %s",
                at.x, at.y, at.z, df3_error_message(ec));
        app_lastec = le2ae(ec);
        return false;
      }
      if (2 == df3i.bpv) {
        uint8_t * const p = pbuf + 2 * icols * y + 2 * x;
        *p = (uint8_t)(v / 256);  *(p + 1) = (uint8_t)(v % 256);
      }
      else
        *(pbuf + icols * y + x) = (uint8_t)v;
    }
  return true;
}

static bool buffer2png(void) {
  if (app_diagnose)
    EMITDMSG(TU, "preparing image data x:%zu y:%zu @ %zu-bits",
            icols, irows, 8 * df3i.bpv);

  bool f;
  if (asrc)
    f = fileLoop();
  else
    f = inputGet(NULL, cback, &df3i);

  if (f)
    f = pngPut(pngname, icols, irows, df3i.bpv, pbuf);

  return f;
}


/* ------------------------------------------------------------------------- */

static void exportTidyup(void) {
  FREENULL(df3i.fname);
  FREENULL(pbuf);
  FREENULL(pngname);
  return;
}

bool exportMain(char *av[]) {
  if (atexit(exportTidyup))
    APPPANIC;

  if (!procArgs(av))
    return false;
  else if (!openSource())
    return false;
  else if (!prepTargetBuffer()) {
    if (!asrc)
      (void)df3_close_file(&df3i);
    return false;
  }

  bool const f = buffer2png();

  if (!(closeFile()  &&  f)) {
    (void)unlink(pngname);
    if (app_diagnose)
      EMITDMSG(TU, "failed target '%s' removed", trimmedFn(pngname));
  }
  else if (app_verbose)
    printf("export from layer %c[%zu] to '%s'.\n",
            aaxis, aoffs, trimmedFn(pngname));

  return f;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
