/* info.c */

/* cf tdestroy(3) */
#define _GNU_SOURCE

#include "df3util.h"

#include <search.h>

static char const * const TU = "info";

static char *asrc;      /* (current) source, null if default */

static df3Axes saxes;   /* axes arg, if source default */
static size_t sbpv;     /* bpv arg, if source default */

static df3Handle df3i;

static uint64_t ndvox,               /* # distinct voxels */
                nzvox,               /* # zero voxels */
                nvvox;               /* # valued voxels */
static uint32_t lovox = UINT32_MAX,  /* lowest/highest voxel value */
                hivox;

typedef struct vCount vCount;  /* infrastructure */
struct vCount {
  uint32_t v;
  uint64_t n;
};

static void *counts;

static int cmpCounts(void const *v1, void const *v2) {
  vCount const * const c1 = (vCount *)v1,
               * const c2 = (vCount *)v2;
  if (c1->v < c2->v)
    return -1;
  else if (c1->v == c2->v)
    return 0;
  else
    return 1;
}

static void resetCounts(void) {
  if (counts) {
    tdestroy(counts, free);
    counts = NULL;
  }
  ndvox = nzvox = nvvox = 0;
  lovox = UINT32_MAX;
  hivox = 0;
  return;
}

static void countVoxel(uint32_t const v) {
  vCount const cmp = (vCount){v, 0};
  vCount *p;
  void *tmp = tfind(&cmp, &counts, cmpCounts);
  if (tmp) {
    p = *(vCount **)tmp;
    p->n += 1;
    return;
  }
  p = malloc(sizeof(vCount));
  if (!p)
    APPPANIC;
  p->v = v;
  p->n = 1;
  tmp = tsearch(p, &counts, cmpCounts);
  if (!tmp)
    APPPANIC;
  return;
}

static void tally(void const *vp, VISIT const which, int const depth AUU) {
  vCount const * const p = *(vCount **)vp;
  switch (which) {
    case postorder:
    case leaf:
      if (p->v  &&  p->v < lovox)
        lovox = p->v;
      if (p->v  &&  p->v > hivox)
        hivox = p->v;
      if (p->v)
        nvvox += p->n;
      else
        nzvox += p->n;  /* '=' would do, however, for symmetry.. */
      ndvox += 1;
      if (app_verbose)
        printf("   %10u   %16lu\n", p->v, p->n);
      break;
    default:
      break;
  }
  return;
}

static void outputInfo(df3Axes const a,
                       size_t const b,
                       char const * const fn) {
  static bool firsttime = true;
  int const w = 17;
  if (firsttime) {
    firsttime = false;
    printf("\n");
  }
  printf("%*s: %s\n", w, "file", trimmedFn(fn));
  printf("%*s: %zu, %zu, %zu\n", w, "volume (x,y,z)", a.x, a.y, a.z);
  printf("%*s: %zu\n", w, "bytes per voxel", b);
  if (app_verbose)
    printf("\n   voxel         count\n  ------------  -----------------\n");
  twalk(counts, tally);
  if (app_verbose)
    printf("\n");
  printf("%*s: %lu\n", w, "distinct values", ndvox);
  printf("%*s: %u, %u\n", w, "range (lo,hi)", nvvox ? lovox : 0, hivox);
  printf("%*s: %lu\n", w, "zero voxels", nzvox);
  printf("%*s: %lu\n", w, "valued voxels", nvvox);
  printf("\n");
  return;
}


/* procArg()
 * -------------------------------------------------------------------------- *
 */

static size_t procArg(char *av[]) {
  char buf[PATH_MAX];
  size_t i;
  size_t const oldsbpv = sbpv;
  int const rc = getInpSrc(av, &saxes, &sbpv);

  if (-1 == rc)
    return 0;
  else if (1 == rc) {
    if (oldsbpv) {
      EMITEMSG(false, "cannot read <stdin> twice");
      app_lastec = ec_badUsage;
      return 0;
    }
    asrc = NULL;
    i = 3;
  }
  else if (!getArgFname(av[0], buf)) {
    EMITEMSG(false, "bad source file name");
    return 0;
  }
  else {
    asrc = av[0];
    i = 1;
  }

  return i;
}


/* procFile()
 * -------------------------------------------------------------------------- *
 * process one actual file.
 */

static bool fileLoop(df3Handle *df3) {
  bool f = false;
  for (size_t z = 0; df3->axes.z != z; ++z)
    for (size_t y = 0; df3->axes.y != y; ++y)
      for (size_t x = 0; df3->axes.x != x; ++x) {
        uint32_t v;
        df3Error ec = df3_get_voxel(df3, (df3Coord){x, y, z}, &v);
        if (ecOk != ec) {
          EMITEMSG(false, "reading voxel <%zu,%zu,%zu>: %s",
                  x, y, z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
        countVoxel(v);
      }

  f = true;

label_loopfail:
  return f;
}

static bool procFile(char * const fname) {
  df3Error ec;

  if (!mkSourceFname(fname, false, false, &df3i.fname)) {
    EMITEMSG(app_warns, "source '%s': %s",
            trimmedFn(asrc), appLastError(app_lastec));
    return false;
  }
  else if (ecOk != (ec = df3_open_file(&df3i, true))) {
    EMITEMSG(app_warns, "open source '%s': %s",
            trimmedFn(fname), df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }
  else if (app_diagnose) {
    EMITDMSG(TU, "source handle");
    df3_handle_info(&df3i, app_logfp);
  }

  bool const f = fileLoop(&df3i);
  if (f)
    outputInfo(df3i.axes, df3i.bpv, df3i.fname);

  ec = df3_close_file(&df3i);

  if (app_diagnose) {
    EMITDMSG(TU, "source handle closed: %s", df3_handle_error(&df3i, false));
    if (!f  ||  ecOk != ec)
      EMITDMSG(TU, "error processing '%s': %s",
              trimmedFn(df3i.fname), df3_error_message(ec));
  }

  return f;
}


/* procInp()
 * -------------------------------------------------------------------------- *
 * wrap callback via inputGet().
 */

static bool cback(void *vp, df3Coord const at, uint32_t const v) {
  df3Handle * const df3 = vp;

  if (!inpVoxelValid(df3->axes, df3->bpv, at, v))
    return false;
    
  countVoxel(v);

  return true;
}

static bool procInp(char * const fname AUU) {
  char const * const fn = "<stdin>";

  df3i.axes = saxes;
  df3i.bpv = sbpv;

  if (app_diagnose)
    EMITDMSG(TU, "source handle %s", fn);

  bool const f = inputGet(NULL, cback, &df3i);
  if (f)
    outputInfo(df3i.axes, df3i.bpv, fn);

  if (app_diagnose)
    EMITDMSG(TU, "end source %s", fn);

  return f;
}


/* ------------------------------------------------------------------------- */

static void infoTidyup(void) {
  if (counts) {
    tdestroy(counts, free);
    counts = NULL;
  }
  FREENULL(df3i.fname);
  return;
}

bool infoMain(char *av[]) {
  if (atexit(infoTidyup))
    APPPANIC;

  bool f = true;
  int nproc = 0, nskip = 0;
  size_t i = 0;

  while (av[i]) {
    size_t const n = procArg(&av[i]);
    if (!n)
      return false;

    f = (asrc ? procFile : procInp)(av[i]);
    if (f)
      nproc += 1;
    else if (app_warns)
      nskip += 1;
    else
      break;
    resetCounts();
    FREENULL(df3i.fname);
    df3i = DF3_INITHANDLE;
    i += n;
  }

  if (f  &&  app_verbose)
    printf("%d file%s processed, %d file%s skipped.\n",
            nproc, 1 == nproc ? "" : "s",
            nskip, 1 == nskip ? "" : "s");

  return f;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
