/* merge.c */

/* cf random(3) */
#define _BSD_SOURCE

#include "df3util.h"

static char const * const TU = "merge";

static bool usextarget;  /* T - open */
                         /* F - create */

static df3Handle df3o;

static df3Axes saxes;    /* axes, if source default */
static size_t sbpv;      /* bpv, if source default */


/* infrastructure */
char const * const mergerules[] = {
  "add", "and", "div", "mid", "mod", "mul",
  "not", "or", "rnd", "set", "sub", "xor",
  NULL
};

enum mRule {
  mrInvalid = -1,
  mrAdd = 0, mrAnd, mrDiv, mrMid, mrMod, mrMul,
  mrNot, mrOr, mrRnd, mrSet, mrSub, mrXor
};
typedef enum mRule mRule;

static uint32_t applyRule(mRule const r,
                          size_t const bpv,
                          uint32_t const v1,
                          uint32_t const v2) {
  int64_t vt = (int64_t)v1;
  int64_t vs = (int64_t)v2;
  switch (r) {
    case mrAdd: vt += vs; break;
    case mrAnd: vt &= vs; break;
    case mrDiv: if (vs) vt /= vs; break;
    case mrMid: vt = (vt + vs) / 2; break;
    case mrMod: if (vs) vt %= vs; break;
    case mrMul: vt *= vs; break;
    case mrNot: vt = ~vs; break;
    case mrOr:  vt |= vs; break;
    case mrRnd: if (vs) vt = random() % (vs + 1); break;
    case mrSet: vt = vs; break;
    case mrSub: vt -= vs; break;
    case mrXor: vt ^= vs; break;
    default:    APPPANIC;
  }
  if (4 == bpv)
    vt &= UINT32_MAX;
  else
    vt &= (2 == bpv) ? UINT16_MAX : UINT8_MAX;
  return (uint32_t)vt;
}

static mRule identifyRule(char const * const s) {
  mRule r = mrInvalid;
  for (size_t i = 0; mergerules[i]; ++i)
    if (!strcmp(s, mergerules[i])) {
      r = (mRule)i;
      break;
    }
  return r;
}


typedef struct mArgs mArgs;
struct mArgs {
  char *asrc;
  df3Axes axes;
  Offset offset;
  mRule rule;
};

#define ZARGS (mArgs){NULL,GENORIGIN,GENORIGIN,mrInvalid}


typedef struct cbData cbData;
struct cbData {
  df3Handle *h;
  df3Axes axes;
  size_t bpv;
  mArgs const *m;
};


/* procArgSet()
 * -------------------------------------------------------------------------- *
 * collect one set of:  source axes offset rule.
 *                 or:  . axes bpv axes offset rule
 */

static size_t procArgSet(char *av[], mArgs *marg) {
  size_t const oldsbpv = sbpv;
  int const rc = getInpSrc(av, &saxes, &sbpv);
  char buf[PATH_MAX];
  size_t i;

  if (-1 == rc)
    return 0;
  else if (1 == rc) {
    if (oldsbpv) {
      EMITEMSG(false, "cannot read <stdin> twice");
      app_lastec = ec_badUsage;
      return 0;
    }
    marg->asrc = NULL;
    i = 3;
  }
  else if (!getArgFname(av[0], buf)) {
    EMITEMSG(false, "bad source file name");
    return 0;
  }
  else {
    marg->asrc = av[0];
    i = 1;
  }

  if (!(av[i]  &&  av[i+1]  &&  av[i+2])) {
    EMITEMSG(false, "missing argument(s)");
    app_lastec = ec_badUsage;
    return 0;
  }

  df3Axes a;
  if (isDefaultArg(av[i]))
    a = (df3Axes)GENORIGIN;
  else if (!getArgCoord(av[i], &a)) {
    EMITEMSG(false, "bad axes value(s)");
    return 0;
  }
  else if (!df3_check_axes(a)) {
    EMITEMSG(false, "zero axes value(s)");
    app_lastec = ec_badAxes;
    return 0;
  }
  marg->axes = a;

  Offset o;
  if (isDefaultArg(av[i+1]))
    o = ZEROOFFS;
  else if (!getArgOffset(av[i+1], &o)) {
    EMITEMSG(false, "bad offset value(s)");
    return 0;
  }
  marg->offset = o;

  mRule r;
  if (isDefaultArg(av[i+2]))
    r = mrSet;
  else if (mrInvalid == (r = identifyRule(av[i+2]))) {
    EMITEMSG(false, "bad merge rule");
    app_lastec = ec_badArgument;
    return 0;
  }
  marg->rule = r;

  /* return # args consumed */
  return 3 + i;
}


/* checkViableTarget()
 * -------------------------------------------------------------------------- *
 * determines whether target already exists or needs creating.
 * returns true (and df3->fname) if all goes well, and sets the boolean
 * which controls the deferred opening.
 */

static bool checkViableTarget(char const * const arg) {

  if (mkSourceFname(arg, false, true, &df3o.fname)) {
    usextarget = true;
    return true;
  }
  else if (mkTargetFname(arg, false, &df3o.fname)) {
    usextarget = false;
    return true;
  }

  EMITEMSG(false, "cannot use target name '%s'", arg);
  app_lastec = ec_badFilename;
  return false;
}


/* openTargetPrep()
 * -------------------------------------------------------------------------- *
 * existing targets are simply opened, new targets, if not too large for disk,
 * get created, and filled if necessary.
 */

static bool openTargetPrep(df3Axes const axes,
                           size_t const bpv,
                           bool const inp) {
  df3Error ec;
  if (usextarget)
    ec = df3_open_file(&df3o, false);
  else if (!haveDiskSpace(axes, bpv))
    ec = ecBadVoxVal;  /* misuse unused (at this point) code */
  else {
    df3o.axes = axes;
    df3o.bpv = bpv;
    ec = df3_create_file(&df3o);
    if ((app_fillval  ||  inp)  &&  ecOk == ec)
      ec = prepFilled(TU, &df3o);
  }
  switch (ec) {
    case ecOk:        break;
    case ecBadVoxVal: app_lastec = ec_diskFull; break;
    default:          app_lastec = le2ae(ec);   break;
  }
  return ecOk == ec;
}


/* mergeInto()
 * -------------------------------------------------------------------------- *
 * on first run, target must be opened or created once the dimensions
 * of the (first) source are known, plus check room when new target.
 * check whether source fits inside target.
 */

static inline uint32_t truncVox(size_t const b, uint32_t v) {
  if (4 != b)
    v &= (2 == b) ? UINT16_MAX : UINT8_MAX;
  return v;
}

static bool cback(void *vp, df3Coord const at, uint32_t const vs) {
  cbData *xfer = vp;
  df3Error ec;
  df3Handle * const df3 = xfer->h;
  uint32_t xvs = vs;
  df3Coord xat = at;

  if (!inpVoxelValid(xfer->axes, xfer->bpv, at, vs))
    return false;
  else if (!containedIn(xfer->m->axes, ZEROOFFS, at))
    return true;
  else if (!atMinters(xfer->m->axes, &xat, xfer->m->offset, df3->axes))
    return true;
  else if (!isValidVoxel(df3->bpv, vs)) {
    EMITEMSG(true, "truncated voxel <%zu,%zu,%zu>", at.x, at.y, at.z);
    xvs = truncVox(df3->bpv, vs);
  }
  uint32_t vt;
  if (ecOk != (ec = df3_get_voxel(df3, xat, &vt))) {
    EMITEMSG(false, "reading target <%zu,%zu,%zu>: %s",
            xat.x, xat.y, xat.z, df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }
  vt = applyRule(xfer->m->rule, df3->bpv, vt, xvs);
  if (ecOk != (ec = df3_put_voxel(df3, xat, vt))) {
    EMITEMSG(false, "writing voxel <%zu,%zu,%zu>: %s",
            xat.x, xat.y, xat.z, df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }
  return true;
}

static bool mergeLoop(mRule const r,
                      df3Axes const a,
                      Offset const o,
                      df3Handle *df3i) {
  bool f = false;
  for (size_t z = 0; a.z != z; ++z)
    for (size_t y = 0; a.y != y; ++y)
      for (size_t x = 0; a.x != x; ++x) {
        df3Error ec;
        df3Coord at, xat;
        uint32_t vt, vs;
        at = xat = (df3Coord){x, y, z};
        if (!atMinters(a, &xat, o, df3o.axes))
          continue;
        else if (ecOk != (ec = df3_get_voxel(df3i, at, &vs))) {
          EMITEMSG(false, "reading source <%zu,%zu,%zu>: %s",
                  at.x, at.y, at.z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
        else if (!isValidVoxel(df3o.bpv, vs)) {
          EMITEMSG(true, "truncated voxel <%zu,%zu,%zu>", at.x, at.y, at.z);
          vs = truncVox(df3o.bpv, vs);
        }

        if (ecOk != (ec = df3_get_voxel(&df3o, xat, &vt))) {
          EMITEMSG(false, "reading target <%zu,%zu,%zu>: %s",
                  xat.x, xat.y, xat.z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
        vt = applyRule(r, df3o.bpv, vt, vs);
        if (ecOk != (ec = df3_put_voxel(&df3o, xat, vt))) {
          EMITEMSG(false, "writing target <%zu,%zu,%zu>: %s",
                  xat.x, xat.y, xat.z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
      }
  f = true;

label_loopfail:
  return f;
}

static bool openSource(df3Handle *df3, mArgs *marg) {
  if (marg->asrc) {
    df3Error ec;
    if (!mkSourceFname(marg->asrc, false, false, &df3->fname)) {
      EMITEMSG(false, "source '%s': %s",
              marg->asrc, appLastError(app_lastec));
      return false;
    }
    else if (ecOk != (ec = df3_open_file(df3, true))) {
      EMITEMSG(false, "open source '%s': %s",
              trimmedFn(df3->fname), df3_error_message(ec));
      app_lastec = le2ae(ec);
      return false;
    }
  }
  else {
    df3->axes = saxes;
    df3->bpv = sbpv;
  }

  if (!df3_check_axes(marg->axes))
    marg->axes = df3->axes;
  else if (df3->axes.x < marg->axes.x
          ||  df3->axes.y < marg->axes.y
          ||  df3->axes.z < marg->axes.z) {
    EMITEMSG(false, "axes larger than source");
    return false;
  }

  return true;
}

static bool mergeInto(mArgs *marg) {
  static bool firsttime = true;

  df3Error ec;
  df3Handle df3i = DF3_INITHANDLE;
  if (!openSource(&df3i, marg))
    return false;

  /* ready target */
  if (firsttime) {
    assert(!df3o.bpv);
    if (!openTargetPrep(df3i.axes, df3i.bpv, marg->asrc ? false : true)) {
      EMITEMSG(false, "preparing target '%s': %s",
              trimmedFn(df3o.fname), appLastError(app_lastec));
      goto label_fail;
    }
    else if (app_diagnose) {
      EMITDMSG(TU,"target handle");
      df3_handle_info(&df3o, app_logfp);
    }
    firsttime = false;
  }

  if (app_diagnose) {
    if (marg->asrc) {
      EMITDMSG(TU, "source handle");
      df3_handle_info(&df3i, app_logfp);
    }
    else
      EMITDMSG(TU, "source handle <stdin>");
  }

  if (df3o.bpv != df3i.bpv) {
    if (!app_warns) {
      EMITEMSG(false, "mismatched bpvs");
      app_lastec = ec_badBPV;
      goto label_fail;
    }
    EMITEMSG(true, "mismatched bpvs%s",
        (df3o.bpv < df3i.bpv) ? ", voxels may be truncated" : "");
  }

  bool junk;
  if (!volsOverlap(df3o.axes, marg->offset, marg->axes, &junk)) {
    EMITEMSG(false, "axes, when offset, not inside target");
    goto label_fail;
  }
  else {
    if (marg->asrc) {
      if (!mergeLoop(marg->rule, marg->axes, marg->offset, &df3i))
        goto label_fail;
      else if (ecOk != (ec = df3_close_file(&df3i))) {
        EMITEMSG(false, "closing source '%s': %s",
                trimmedFn(df3i.fname), df3_error_message(ec));
        app_lastec = le2ae(ec);
        FREENULL(df3i.fname);
        return false;
      }
    }
    else {
      cbData x;
      x.h = &df3o;
      x.axes = df3i.axes;
      x.bpv = df3i.bpv;
      x.m = marg;
      if (!inputGet(NULL, cback, &x))
        goto label_fail;
    }
  }

  if (app_diagnose) {
    if (marg->asrc)
      EMITDMSG(TU, "source handle closed: %s", df3_handle_error(&df3i, false));
    else
      EMITDMSG(TU, "end source <stdin>");
  }

  FREENULL(df3i.fname);
  return true;

label_fail:
  if (df3i.fname)
    ec = df3_close_file(&df3i);
// TODO update app_lastec if failed?
  if (app_diagnose  &&  df3i.fname)
    EMITDMSG(TU, "source handle closed: %s",
            df3_handle_error(&df3i, false));
  FREENULL(df3i.fname);
  return false;
}


/* ------------------------------------------------------------------------- */

static void mergeTidyup(void) {
  FREENULL(df3o.fname);
}

bool mergeMain(char *av[]) {
  if (atexit(mergeTidyup))
    APPPANIC;

  if (!checkViableTarget(av[0]))
    return false;

  bool f = true;
  int nproc = 0;
  mArgs args = ZARGS;
  size_t i = 1;
  while (av[i]) {
    size_t const n = procArgSet(&av[i], &args);
    if (!n) {
      f = false;
      break;
    }
    else if (!mergeInto(&args)) {
      f = false;
      break;
    }
    nproc += 1;
    args = ZARGS;
    i += n;
  }

  if (df3o.bpv) {
    df3Error const ec = df3_close_file(&df3o);
    if (app_diagnose)
      EMITDMSG(TU, "target handle closed: %s",
              df3_handle_error(&df3o, false));
    if (f  &&  ecOk == ec) {
      if (app_verbose) {
        printf("merge into %s target '%s',\n",
                usextarget ? "existing" : "new", trimmedFn(df3o.fname));
        printf("%d source file%s processed.\n", nproc, 1 == nproc ? "" : "s");
      }
    }
    else if (!f) {
      if (usextarget)
        EMITEMSG(false, "failed merge into existing target '%s',"
                        " data (perhaps) corrupted",
                trimmedFn(df3o.fname));
      else {
        (void)unlink(df3o.fname);
        EMITEMSG(false, "failed merge into new target '%s',"
                        " target removed",
                trimmedFn(df3o.fname));
      }
    }
  }

  return f;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
