/* pngrw.c */

#include "df3util.h"

#include <png.h>
#include <time.h>


static char const * const TU = "png";

/* cf signature check */
#define SIGLEN 7

/* keep (shared) stuff in one place */
typedef struct pInfo pInfo;
struct pInfo {
  png_structp png;     /* the mandatory i/f */
  png_infop inf1,
            inf2;
  char const *fname;   /* supplied */
  FILE *fp;
  size_t bpv;          /* supplied */
  uint32_t cols,       /* supplied or read */
           rows;
  double gamma;        /* read */
  png_byte bitdepth;   /* supplied or read */
  int channels,        /* read */
      colourtype,      /* supplied or read */
      interlace;       /* supplied or read */
  bool alpha;          /* read */
  png_color_8 sbits;   /* supplied or read */
  png_time datim;      /* supplied or read */
};

static pInfo I;


/* prettyInfo()
 * -------------------------------------------------------------------------- *
 * TODO expand to list chunk name/counts/select content?
 * TODO need two parts, cf png_read_end().
 */

static char const *prettyColour(int const c) {
  static char desc[40];
  sprintf(desc, "%spalette, %s, %salpha",
          (PNG_COLOR_MASK_PALETTE & c) ? "" : "no ",
          (PNG_COLOR_MASK_COLOR & c) ? "rgb" : "grayscale",
          (PNG_COLOR_MASK_ALPHA & c) ? "" : "no ");
  return desc;
}

static char const *prettyChannels(int const c) {
  static char desc[40];
  switch (c) {
    case 1:   strcpy(desc, "palette or gray");  break;
    case 2:   strcpy(desc, "gray alpha");  break;
    case 3:   strcpy(desc, "rgb");  break;
    case 4:   strcpy(desc, "rgb alpha");  break;
    default:  strcpy(desc, "(invalid)");  break;
  }
  return desc;
}

static char const *prettyDatim(png_time const dtm) {
  static char desc[40];
  if (dtm.year)
    sprintf(desc, "%d.%d.%d %02d:%02d.%02d",
            dtm.day, dtm.month, dtm.year,
            dtm.hour, dtm.minute, dtm.second);
  else
    strcpy(desc, "n/a");
  return desc;
}

static void prettyInfo(pInfo const * const p) {
  int const w = 16;
  EMITDMSG(TU, "source meta");
  fprintf(app_logfp, "%*s: %s\n"
                     "%*s: %s\n"
                     "%*s: %u x %u\n"
                     "%*s: %s\n"
                     "%*s: %s\n"
                     "%*s: %d - %s\n"
                     "%*s: %d\n"
                     "%*s: %d.%d.%d.%d.%d\n"
                     "%*s: %f\n",
      w, "filename", trimmedFn(p->fname),
      w, "last edit", prettyDatim(p->datim),
      w, "cols x rows", p->cols, p->rows,
      w, "interlaced", PNG_INTERLACE_NONE == p->interlace ? "no" : "yes",
      w, "colour type", prettyColour(p->colourtype),
      w, "# channels", p->channels, prettyChannels(p->channels),
      w, "bits/channel", p->bitdepth,
      w, "signif.(rgbGa)", p->sbits.red, p->sbits.green, p->sbits.blue,
              p->sbits.gray, p->sbits.alpha,
      w, "gamma", p->gamma);
  return;
}


/* transforms()
 * -------------------------------------------------------------------------- *
 * expect 8-bit or 16-bit grayscale.  for now drop alpha.  if {app_warns}
 * also unpack 1,2,4-bit and convert rgb to gray.
 * TODO libpng's 'example.c' mentions that if sRGB present meaning of other
 * chunks (gamma+alpha) changes.  need info/examples.
 */

static bool transforms(pInfo *p) {
  bool const isgray = (PNG_COLOR_TYPE_GRAY_ALPHA == p->colourtype
                    || PNG_COLOR_TYPE_GRAY == p->colourtype);

  if (p->alpha)
    png_set_strip_alpha(p->png);

  if (2 == p->bpv) {
    if (16 != p->bitdepth) {
      EMITEMSG(false, "bpv bit depth mismatch");
      return false;
    }
    else if (!isgray) {
      if (!app_warns) {
        EMITEMSG(false, "image not grayscale");
        return false;
      }
      png_set_rgb_to_gray_fixed(p->png, 1, -1, -1);
      if (app_diagnose)
        EMITDMSG(TU, "convert to grayscale");
    }
  }
  else if (16 == p->bitdepth) {
    EMITEMSG(false, "bpv bit depth mismatch");
    return false;
  }
  else { /* 1 == bpv AND 8 >= depth */
    if (!(isgray  ||  app_warns)) {
        EMITEMSG(false, "image not grayscale");
        return false;
    }
    if (PNG_COLOR_TYPE_PALETTE == p->colourtype  &&  8 >= p->bitdepth) {
      png_set_palette_to_rgb(p->png);
      if (app_diagnose)
        EMITDMSG(TU, "convert palette to rgb");
    }
    if (PNG_COLOR_TYPE_GRAY == p->colourtype  &&  8 >= p->bitdepth) {
      png_set_expand(p->png);
      if (app_diagnose)
        EMITDMSG(TU, "convert %d-bit gray to 8-bit", p->bitdepth);
    }
    if (!isgray) {
      png_set_rgb_to_gray_fixed(p->png, 1, -1, -1);
      if (app_diagnose)
        EMITDMSG(TU, "convert to grayscale");
    }
  }

  png_set_gamma(p->png, app_gamma, p->gamma);

  png_read_update_info(p->png, p->inf1);

  return true;
}


/* sizeCheck()
 * -------------------------------------------------------------------------- *
 * image cols x rows may be smaller than target width x height, {app_warns}
 * permitting, but not zero nor larger.
 */

static bool sizeCheck(size_t const w,
                      size_t const h,
                      png_uint_32 const c,
                      png_uint_32 const r) {
  char const * const fmt =
          "mismatched dimensions (%zux%zu buffer vs %ux%u file)";
  size_t const x = (size_t)c, y = (size_t)r;
  bool f;
  if (!(x * y)  ||  w < x  ||  h < y) {
label_emsg:
    app_lastec = ec_badArgument;
    EMITEMSG(false, fmt, w, h, c, r);
    f = false;
  }
  else if (w == x  &&  h == y)
    f = true;
  else if (!app_warns)
    goto label_emsg;
  else {
    EMITEMSG(true, fmt, w, h, c, r);
    f = true;
  }
  return f;
}


/* mkStructs()
 * -------------------------------------------------------------------------- *
 * required infrastructure.
 */

static bool mkStructs(pInfo *p, bool const ro) {
  char const * const fmt = "libpng %s: %s";

  if (ro) {
    p->png = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!p->png) {
      int const e = errno;
      app_lastec = ec_sysError;
      EMITEMSG(false, fmt, "create_read_struct", strerror(e));
      return false;
    }
  }
  else {
    p->png = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!p->png) {
      int const e = errno;
      app_lastec = ec_sysError;
      EMITEMSG(false, fmt, "create_write_struct", strerror(e));
      return false;
    }
  }

  p->inf1 = png_create_info_struct(p->png);
  if (!p->inf1) {
    int const e = errno;
    app_lastec = ec_sysError;
    EMITEMSG(false, fmt, "create_info_struct", strerror(e));
    if (ro)
      png_destroy_read_struct(&p->png, NULL, NULL);
    else
      png_destroy_write_struct(&p->png, NULL);
    return false;
  }

  if (ro) {
    p->inf2 = png_create_info_struct(p->png);
    if (!p->inf2) {
      int const e = errno;
      app_lastec = ec_sysError;
      EMITEMSG(false, fmt, "create_info_struct", strerror(e));
      png_destroy_read_struct(&p->png, &p->inf1, NULL);
      return false;
    }
  }

  return true;
}


/* openPng()
 * -------------------------------------------------------------------------- *
 * regular file known to exist + readable.
 */

static bool openPng(char const * const fname, FILE **pfp) {
  png_byte signature[SIGLEN];

  FILE *fp = fopen(fname, "rb");
  if (!fp) {
    int const e = errno;
    app_lastec = ec_sysError;
    EMITEMSG(false, "opening source: %s", strerror(e));
  }
  else if (SIGLEN != fread(signature, sizeof(png_byte), SIGLEN, fp)) {
    int const e = errno;
    app_lastec = ec_sysError;
    EMITEMSG(false, "reading source: %s", strerror(e));
  }
  else if (png_sig_cmp(signature, 0, SIGLEN)) {
    app_lastec = ec_noFile;
    EMITEMSG(false, "source '%s' not PNG", trimmedFn(fname));
  }
  else if (fseek(fp, 0, SEEK_SET)) {
    int const e = errno;
    app_lastec = ec_sysError;
    EMITEMSG(false, "source seek: %s", strerror(e));
  }
  else {
    *pfp = fp;
    if (app_diagnose)
      EMITDMSG(TU, "opened PNG source '%s'", trimmedFn(fname));
    return true;
  }

  (void)fclose(fp);
  return false;
}


/* procData()
 * -------------------------------------------------------------------------- *
 * read image data into {buf}.
 */

static bool procData(pInfo *p,
                     size_t const w,
                     size_t const h,
                     void *buf,
                     int *ecode) {
  size_t const rowbytes = png_get_rowbytes(p->png, p->inf1);
  bool f = false;
  png_bytep *rows;
  if (!(rows = calloc(p->rows, sizeof(png_bytep)))) {
    *ecode = errno;
    return f;
  }
  for (size_t i = 0; p->rows != i; ++i)
    if (!(rows[i] = calloc(rowbytes, sizeof(png_byte)))) {
      *ecode = errno;
      goto label_fail;
    }

  png_read_image(p->png, rows);

  for (size_t i = 0; h != i; ++i)
    if (p->rows > i)
      for (size_t j = 0; w != j; ++j)
        if (rowbytes > j) {
          if (2 == p->bpv)
            *((uint16_t *)buf + w * i + j) =
                (uint16_t)(*(rows[i] + 2 * j) * 256 + *(rows[i] + 2 * j + 1));
          else
            *((uint8_t *)buf + w * i + j) = (uint8_t)*(rows[i] + j);
        }

  f = true;

label_fail:
  for (size_t i = 0; p->rows != i; ++i)
    if (rows[i])
      FREENULL(rows[i]);
  FREENULL(rows);
  return f;
}


/* infoRead()
 * -------------------------------------------------------------------------- *
 * read image header and stuff.
 */

static void infoRead(pInfo *p) {

  png_read_info(p->png, p->inf1);

  int tbitdepth = 0;
  png_get_IHDR(p->png, p->inf1,
          &p->cols, &p->rows,
          &tbitdepth, &p->colourtype, &p->interlace,
          NULL, NULL);
  p->bitdepth = (png_byte)tbitdepth;

  p->channels = png_get_channels(p->png, p->inf1);

  if (png_get_valid(p->png, p->inf1, PNG_INFO_tRNS))
    p->alpha = true;
  else
    p->alpha = (PNG_COLOR_MASK_ALPHA & p->colourtype);

  if (png_get_valid(p->png, p->inf1, PNG_INFO_gAMA))
    png_get_gAMA(p->png, p->inf1, &p->gamma);
  else
    p->gamma = 0.45;  /* TODO leave at zero? */

  if (png_get_valid(p->png, p->inf1, PNG_INFO_sBIT)) {
    png_color_8p tc = NULL;
    png_get_sBIT(p->png, p->inf1, &tc);
    p->sbits = *tc;
  }

  if (png_get_valid(p->png, p->inf1, PNG_INFO_tIME)) {
    png_timep ts = NULL;
    png_get_tIME(p->png, p->inf1, &ts);
    p->datim = *ts;
  }

  return;
}


/* pngGet()
 * -------------------------------------------------------------------------- *
 * read grayscale image data into {buf}.  {fname} exists and is readable,
 * should be 8|16 bit grayscale, with '-warnings' accept 1|2|4 grayscales and
 * rgb or rgb/a; use formula in povray manual 3.4.1.4 HF_Gray_16, 
 */

bool pngGet(char const * const fname,
            size_t const width,
            size_t const height,
            size_t const bpv,
            void *buf) {

  errno = 0;
  I.fname = fname;
  I.bpv = bpv;

  if (!mkStructs(&I, true))
    return false;
  else if (!openPng(fname, &I.fp)) {
    png_destroy_read_struct(&I.png, &I.inf1, &I.inf2);
    return false;
  }

  if (setjmp(png_jmpbuf(I.png))) {
label_fail:
    png_destroy_read_struct(&I.png, &I.inf1, &I.inf2);
    (void)fclose(I.fp);
    return false;
  }

  png_init_io(I.png, I.fp);

  infoRead(&I);

  if (app_diagnose)
    prettyInfo(&I);

  int ecode = 0;
  if (!sizeCheck(width, height, I.cols, I.rows))
    goto label_fail;
  else if (!transforms(&I))
    goto label_fail;
  else if (!procData(&I, width, height, buf, &ecode)) {
    EMITEMSG(false, "%s", strerror(ecode));
    app_lastec = ec_sysError;
    goto label_fail;
  }

  png_read_end(I.png, I.inf2);

// TODO -diag text + unknown chunks.

  png_destroy_read_struct(&I.png, &I.inf1, &I.inf2);
  if(fclose(I.fp))
    EMITEMSG(false, "closing source '%s'", trimmedFn(fname));

  return true;
}



/* infoWrite()
 * -------------------------------------------------------------------------- *
 * provide info and actuate.  cf libpng's 'example.c'.
 * use predefined key 'Software'.
 */

static void infoWrite(pInfo *p) {

  png_set_IHDR(p->png, p->inf1,
          p->cols, p->rows,
          p->bitdepth,
          PNG_COLOR_TYPE_GRAY,
          PNG_INTERLACE_NONE,
          PNG_COMPRESSION_TYPE_BASE,
          PNG_FILTER_TYPE_BASE);

  png_set_sBIT(p->png, p->inf1, &p->sbits);

  png_set_gAMA(p->png, p->inf1, 0.45455);

  char ident[40];
  png_text txt[1];
  sprintf(ident, "%s %s", app_name, app_version);
  txt[0] = (png_text){.compression = -1,
                      .key = (char *)"Software",
                      .text = ident,
                      .text_length = strlen(ident),
                      .itxt_length = 0,
                      .lang = NULL,
                      .lang_key = NULL};
  png_set_text(p->png, p->inf1, txt, 1);

  png_set_oFFs(p->png, p->inf1, 0, 0, PNG_OFFSET_PIXEL);

  time_t t = time(NULL);
  struct tm *now = gmtime(&t);
  p->datim = (png_time){.year = (png_uint_16)(1900 + now->tm_year),
                        .month = (png_byte)(1 + now->tm_mon),
                        .day = (png_byte)now->tm_mday,
                        .hour = (png_byte)now->tm_hour,
                        .minute = (png_byte)now->tm_min,
                        .second = (png_byte)now->tm_sec};
  png_set_tIME(p->png, p->inf1, &p->datim);

  /* .gray=0 ought to be black */
  png_color_16 bgc = {0,0,0,0,0};
  png_set_bKGD(p->png, p->inf1, &bgc);

  png_write_info(p->png, p->inf1);

  return;
}


/* pngPut()
 * -------------------------------------------------------------------------- *
 * create grayscale image data file from {buf}.
 */

bool pngPut(char const * const fname,
            size_t const width,
            size_t const height,
            size_t const bpv,
            void *buf) {

  errno = 0;
  I.fname = fname;
  I.cols = (uint32_t)width;
  I.rows = (uint32_t)height;
  I.bpv = bpv;
  I.bitdepth = (png_byte)(8 * bpv);
  I.sbits.gray = I.bitdepth;
  I.fp = fopen(fname, "wb");

  if (!I.fp) {
    int const e = errno;
    EMITEMSG(false, "opening source '%s': %s", trimmedFn(fname), strerror(e));
    app_lastec = ec_sysError;
    return false;
  }
  else if (!mkStructs(&I, false))
    goto label_fail_nostruct;

  if (setjmp(png_jmpbuf(I.png))) {
label_fail:
    png_destroy_write_struct(&I.png, &I.inf1);
label_fail_nostruct:
    (void)fclose(I.fp);
    (void)unlink(I.fname);
    if (app_diagnose)
      EMITDMSG(TU, "failed target '%s' removed", trimmedFn(I.fname));
    return false;
  }

  png_init_io(I.png, I.fp);

  infoWrite(&I);

// TODO -diag pretty print?

  png_bytep *rows = calloc(height, sizeof(png_bytep));
  if (!rows) {
    int const e = errno;
    EMITEMSG(false, "allocating row pointers: %s", strerror(e));
    goto label_fail;
  }
  for (size_t i = 0; height != i; ++i)
    rows[i] = (png_bytep)buf + width * bpv * i;

  png_write_image(I.png, rows);

  png_write_end(I.png, I.inf1);

// TODO -diag?

  if (fclose(I.fp))
    goto label_fail;

  png_destroy_write_struct(&I.png, &I.inf1);
  FREENULL(rows);

  return true;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
