/* rotate.c */

#include "df3util.h"

static char const * const TU = "rotate";

/* globals */
static char *asrc,      /* source name arg, null if default */
            *atgt;      /* target name arg */
static df3Axes arv;     /* rvect angles arg */

static df3Axes saxes;   /* source axes arg, if source default */
static size_t sbpv;     /* source bpv arg, if source default */

static df3Handle df3i, df3o;

/* re-orient volume, ignore sign change */
static df3Coord rotVol(df3Coord sv, df3Coord const rv) {
  switch (rv.x) {
    case 0:  case 180: break;
    case 90: case 270: sv = (df3Coord){sv.x, sv.z, sv.y}; break;
    default:           APPPANIC;
  }
  switch (rv.y) {
    case 0:  case 180: break;
    case 90: case 270: sv = (df3Coord){sv.z, sv.y, sv.x}; break;
    default:           APPPANIC;
  }
  switch (rv.z) {
    case 0:  case 180: break;
    case 90: case 270: sv = (df3Coord){sv.y, sv.x, sv.z}; break;
    default:           APPPANIC;
  }
  return sv;
}

/* rotate one voxel */
static df3Coord rotVoxelX(df3Coord *dim, df3Coord const at, size_t const r) {
  int y = 0, z = 0;

  switch (r) {
    case 0:
      y = (int)at.y;
      z = (int)at.z;
      break;
    case 90:
      y = -((int)at.z + 1) + (int)dim->z;
      z = (int)at.y;
      break;
    case 180:
      y = -((int)at.y + 1) + (int)dim->y;
      z = -((int)at.z + 1) + (int)dim->z;
      break;
    case 270:
      y = (int)at.z;
      z = -((int)at.y + 1) + (int)dim->y;
      break;
    default:  APPPANIC;
  }

  *dim = rotVol(*dim, (df3Coord){r, 0, 0});
  assert(0 <= y  &&  y < (int)dim->y);
  assert(0 <= z  &&  z < (int)dim->z);
  return (df3Coord){at.x, (size_t)y, (size_t)z};
}

static df3Coord rotVoxelY(df3Coord *dim, df3Coord const at, size_t const r) {
  int x = 0, z = 0;

  switch (r) {
    case 0:
      x = (int)at.x;
      z = (int)at.z;
      break;
    case 90:
      x = (int)at.z;
      z = -((int)at.x + 1) + (int)dim->x;
      break;
    case 180:
      x = -((int)at.x + 1) + (int)dim->x;
      z = -((int)at.z + 1) + (int)dim->z;
      break;
    case 270:
      x = -((int)at.z + 1) + (int)dim->z;
      z = (int)at.x;
      break;
    default:  APPPANIC;
  }

  *dim = rotVol(*dim, (df3Coord){0, r, 0});
  assert(0 <= x  &&  x < (int)dim->x);
  assert(0 <= z  &&  z < (int)dim->z);
  return (df3Coord){(size_t)x, at.y, (size_t)z};
}

static df3Coord rotVoxelZ(df3Coord *dim, df3Coord const at, size_t const r) {
  int x = 0, y = 0;

  switch (r) {
    case 0:
      x = (int)at.x;
      y = (int)at.y;
      break;
    case 90:
      x = -((int)at.y + 1) + (int)dim->y;
      y = (int)at.x;
      break;
    case 180:
      x = -((int)at.x + 1) + (int)dim->x;
      y = -((int)at.y + 1) + (int)dim->y;
      break;
    case 270:
      x = (int)at.y;
      y = -((int)at.x + 1) + (int)dim->x;
      break;
    default:  APPPANIC;
  }

  *dim = rotVol(*dim, (df3Coord){0, 0, r});
  assert(0 <= x  &&  x < (int)dim->x);
  assert(0 <= y  &&  y < (int)dim->y);
  return (df3Coord){(size_t)x, (size_t)y, at.z};
}

static df3Coord rotVox(df3Coord const dim, df3Coord at, df3Coord const rv) {
  df3Coord tmp = dim;
  assert(at.x < dim.x  &&  at.y < dim.y  &&  at.z < dim.z);
  at = rotVoxelX(&tmp, at, rv.x);
  at = rotVoxelY(&tmp, at, rv.y);
  at = rotVoxelZ(&tmp, at, rv.z);
  return at;
}


/* procArgs()
 * -------------------------------------------------------------------------- *
 * cli := source rvect target.
 *     := . axes bpv rvect target.
 */

static bool goodRvect(df3Coord const rv) {
  bool f = true;

  switch (rv.x) {
    case 0: case 90: case 180: case 270:
      break;
    default:
      f = false;
      break;
  }
  switch (rv.y) {
    case 0: case 90: case 180: case 270:
      break;
    default:
      f = false;
      break;
  }
  switch (rv.z) {
    case 0: case 90: case 180: case 270:
      break;
    default:
      f = false;
      break;
  }

  if (!f) {
    app_lastec = ec_badArgument;
    EMITEMSG(false, "bad rotate angle");
  }

  return f;
}

static bool procArgs(char *av[]) {
  bool f = false;
  char buf[PATH_MAX];
  int const rc = getInpSrc(av, &saxes, &sbpv);
  int i;

  if (-1 == rc)
    goto label_fail;
  else if (1 == rc)
    i = 3;
  else if (!getArgFname(av[0], buf)) {
    EMITEMSG(false, "bad source file name");
    goto label_fail;
  }
  else {
    asrc = av[0];
    i = 1;
  }

  if (!(av[i]  &&  av[i+1])) {
    EMITEMSG(false, "missing argument(s)");
    app_lastec = ec_badUsage;
    goto label_fail;
  }

  df3Coord x = GENORIGIN;
  if (!getArgCoord(av[i], &x)) {
    EMITEMSG(false, "bad rvect argument");
    goto label_fail;
  }
  else if (!goodRvect(x))
    goto label_fail;
  arv = x;

  if (!getArgFname(av[i+1], buf)) {
    EMITEMSG(false, "bad target file name");
    goto label_fail;
  }
  atgt = av[i+1];

  f = true;

label_fail:
  return f;
}


/* openSource()
 * -------------------------------------------------------------------------- *
 * open if actual file, pre-compute addend.
 */


static bool openSource(void) {

  if (asrc) {
    df3Error ec;
    if (!mkSourceFname(asrc, false, false, &df3i.fname)) {
      EMITEMSG(false, "source '%s': %s",
              trimmedFn(asrc), appLastError(app_lastec));
      return false;
    }
    else if (ecOk != (ec = df3_open_file(&df3i, true))) {
      EMITEMSG(false, "open source '%s': %s",
              trimmedFn(df3i.fname), df3_error_message(ec));
      app_lastec = le2ae(ec);
      return false;
    }
  }
  else {
    df3i.axes = saxes;
    df3i.bpv = sbpv;
  }

  if (app_diagnose) {
    if (asrc) {
      EMITDMSG(TU, "source handle");
      df3_handle_info(&df3i, app_logfp);
    }
    else
      EMITDMSG(TU, "source handle <stdin>");
  }

  return true;
}


/* openTarget()
 * -------------------------------------------------------------------------- *
 * create file re-oriented, fill if source stdin.
 */

static bool openTarget(void) {

  if (!mkTargetFname(atgt, false, &df3o.fname)) {
    EMITEMSG(false, "target '%s': %s",
            trimmedFn(atgt), appLastError(app_lastec));
    return false;
  }
  else if (!haveDiskSpace(df3i.axes, df3i.bpv)) {
    EMITEMSG(false, "target '%s': %s",
            trimmedFn(df3o.fname), appLastError(app_lastec = ec_diskFull));
    return false;
  }

  df3o.axes = rotVol(df3i.axes, arv);
  df3o.bpv = df3i.bpv;

  df3Error ec = df3_create_file(&df3o);
  if (ecOk != ec) {
    EMITEMSG(false, "create target '%s': %s",
            trimmedFn(df3o.fname), df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }
  else if (app_diagnose) {
    EMITDMSG(TU, "target handle");
    df3_handle_info(&df3o, app_logfp);
  }

  if (!asrc)
    if (!prepFilled(TU, &df3o)) {
      (void)df3_close_file(&df3o);
      return false;
    }

  return true;
}


/* closeFiles()
 * -------------------------------------------------------------------------- *
 * TODO what's lost by setting {app_lastec} here?
 * TODO use EMITEMSGs instead + unconditionally?
 */

static bool closeFiles(void) {
  bool f = false;
  df3Error ec;

  if (asrc) {
    if (ecOk != (ec = df3_close_file(&df3i))) {
      if (app_diagnose)
        EMITDMSG(TU, "close source: %s", df3_error_message(ec));
    }
    else if (app_diagnose)
      EMITDMSG(TU, "closed source handle: %s",
              df3_handle_error(&df3i, false));
    if (ec_sysError != app_lastec)
      app_lastec = le2ae(ec);
  }
  else {  /* redundant, but hey.. */
    (void)fclose(stdin);
    if (app_diagnose)
      EMITDMSG(TU, "closed source <stdin>");
  }

  ec = df3_close_file(&df3o);
  if (ecOk != ec) { 
    if (app_diagnose)
      EMITDMSG(TU, "close target: %s", df3_error_message(ec));
    if (ec_sysError != app_lastec)
      app_lastec = le2ae(ec);
  }
  else {
    if (app_diagnose)
      EMITDMSG(TU, "closed target handle: %s", df3_handle_error(&df3o, false));
    f = true;
  }

  return f;
}


/* procInp()
 * -------------------------------------------------------------------------- *
 * wraps callback via inputGet().
 */

static bool cback(void *vp, df3Coord const at, uint32_t const v) {
  df3Error ec;
  df3Handle * const df3 = vp;

  if (!inpVoxelValid(df3i.axes, df3i.bpv, at, v))
    return false;
  else if (ecOk != (ec = df3_put_voxel(df3, rotVox(df3i.axes, at, arv), v))) {
    EMITEMSG(false, "writing voxel <%zu,%zu,%zu>: %s",
            at.x, at.y, at.z, df3_error_message(ec));
    app_lastec = le2ae(ec);
    return false;
  }

  return true;
}

static inline bool procInp(void) {
  return inputGet(NULL, cback, &df3o);
}


/* procFile()
 * -------------------------------------------------------------------------- *
 */

static bool procFile(void) {
  bool f = false;

  for (size_t z = 0; df3i.axes.z != z; ++z)
    for (size_t y = 0; df3i.axes.y != y; ++y)
      for (size_t x = 0; df3i.axes.x != x; ++x) {
        df3Error ec;
        uint32_t v;
        df3Coord at = {x, y, z};
        if (ecOk != (ec = df3_get_voxel(&df3i, at, &v))) {
          EMITEMSG(false, "reading voxel <%zu,%zu,%zu>: %s",
                  at.x, at.y, at.z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
        at = rotVox(df3i.axes, at, arv);
        if (ecOk != (ec = df3_put_voxel(&df3o, at, v))) {
          EMITEMSG(false, "writing voxel <%zu,%zu,%zu>: %s",
                  at.x, at.y, at.z, df3_error_message(ec));
          app_lastec = le2ae(ec);
          goto label_loopfail;
        }
      }

  f = true;

label_loopfail:
  return f;
}


/* ------------------------------------------------------------------------- */

static void rotateTidyup(void) {
  FREENULL(df3i.fname);
  FREENULL(df3o.fname);
  return;
}

bool rotateMain(char *av[]) {
  if (atexit(rotateTidyup))
    APPPANIC;

  if (!procArgs(av))
    return false;
  else if (!openSource())
    return false;
  else if (!openTarget()) {
    (void)df3_close_file(&df3i);
    return false;
  }

  bool const f = (asrc ? procFile : procInp)();

  if (!(closeFiles()  &&  f)) {
    (void)unlink(df3o.fname);
    if (app_diagnose)
      EMITDMSG(TU, "failed target '%s' removed", trimmedFn(df3o.fname));
  }
  else if (app_verbose)
    printf("volume rotated by <%zu,%zu,%zu>.\n", arv.x, arv.y, arv.z);

  return f;
}

/* ------------------------------------------------------------------ *
 *  this file is part of 'df3util', part of the 'df3 tools' project.  *
 *  the content above is covered by the GPLv3+, see file COPYING.     *
 *  copyright (c) 2017 jr <creature.eternal@gmail.com>.               *
 *  all rights reserved.                                              *
 * ------------------------------------------------------------------ */
